/*
Compilar con:
sdcc -mz80 --opt-code-speed --nostdlib --nostdinc --no-std-crt0 --out-fmt-s19 --code-loc XXXX --data-loc 0 fire_sdcc.c "C:\Archivos de programa\SDCC\lib\z80\mul.o"
Donde XXXX es la direccin de comienzo (p. ej: 32768)
*/

typedef unsigned char BYTE;
typedef unsigned short WORD;

enum {IBLACK=0, IBLUE, IRED, IMAG, IGREEN, ICYAN, IYELLOW, IWHITE};
enum {PBLACK=0, PBLUE=8, PRED=16, PMAG=24, PGREEN=32, PCYAN=40, PYELLOW=48, PWHITE=56};
#define BRIGHT 64
#define FLASH 128

__sfr __at (0xfe) ULA;
__sfr __at (0xff) ATTR;
__sfr __banked __at (0xf7fe) SEMIFILA1;
__sfr __banked __at (0xeffe) SEMIFILA2;
__sfr __banked __at (0xfbfe) SEMIFILA3;
__sfr __banked __at (0xdffe) SEMIFILA4;
__sfr __banked __at (0xfdfe) SEMIFILA5;
__sfr __banked __at (0xbffe) SEMIFILA6;
__sfr __banked __at (0xfefe) SEMIFILA7;
__sfr __banked __at (0x7ffe) SEMIFILA8;

#define COORDS 23296
#define SPOSN 23298
#define ATTRP 23300
#define BORDR 23301
#define CHARS 23606

#define COLUMN (SPOSN)
#define ROW (SPOSN+1)

#define WAIT_VRETRACE _asm halt _endasm
#define WAIT_HRETRACE while(ATTR!=0xff)
#define SETCOLOR(x) *(BYTE *)(ATTRP)=(x)

#pragma preproc_asm -
#define INIT_ZX _asm
  ld hl,32768
  _endasm
#pragma preproc_asm +

void cls (BYTE);
void border (BYTE);
void locate (BYTE, BYTE);

void puts (BYTE *);
void putdec (WORD);
void u16todec (WORD, char *);
void u16tohex (WORD, char *);
void u8tohex (BYTE, char *);

void memset (BYTE *, BYTE, WORD);
void memcpy (BYTE *, BYTE *, WORD);

int abs (int);
signed char sgn (int);

long frames (void) __naked;
void pause (BYTE);

void wait_key (void);
BYTE inkey (BYTE, BYTE);

void beep (WORD, BYTE) __critical;

/* --------------------------------------------------------------------------------- */
/* --------------------------------------------------------------------------------- */
/* --------------------------------------------------------------------------------- */
#define RETARDO_TOTAL 300

//En estas matrices guardamos las posiciones de pantalla (fila,col) de los actores en escena
BYTE posbombero[3][2];  //3 posibles posiciones de los bomberos
BYTE posvictimas[23][2]; //23 posibles posiciones de una victima
BYTE posangelitos[3][2]; //3 indicadores de vida
BYTE poschocazos[3][2]; //3 posiciones para chocazos contra el suelo

//Vector que nos dice con qu frecuencia van apareciendo nuevas victimas en el balcn
BYTE siguiente[132];

//Originalmente era un vector de estructuras pero el SDCC parece que no le hace mucha gracia.
//Hay provisin para hasta 10 victimas en escena a la vez.
BYTE VictimaActiva[10]; // Si la victima i est activa (se est usando esta entrada)
BYTE VictimaPosicion[10]; // Posicion de la victima dentro de posvictimas
BYTE VictimaTick[10]; // En qu momento se actualiza la posicion de esta victima
BYTE VictimaSalvada[10]; // Indica que la victima ha llegado a tocar la lona y no debe chocarse contra el suelo

void presentacion (void);
void init_victimas (void);
void init_matrices (void);
void dibujar_bombero (BYTE pos, BYTE color);
void dibujar_victima (BYTE pos, BYTE color);
void dibujar_angelito (BYTE pos, BYTE color);
void poner_puntos (WORD puntos);
void pegarse_chocazo (BYTE poschoc, BYTE pos_victima, BYTE *vidas);
void borrar_sprites (void);

void main (void)
{
  BYTE pos_bombero,bascula_tiempo,tick,marcatiempo,vidas_restantes,sig_actor,ind_sig_actor,actor,pos_victima,primera_victima_pintada,esta_salvada;
  BYTE izq_pres,der_pres;
  WORD puntos;
  WORD retardo_victimas;
  WORD retardo_maximo;

  SETCOLOR(BRIGHT|PGREEN|IBLACK);
  border(1);

  init_matrices();
  presentacion();

  while (1)  // el bucle de partida: en cada vuelta, se comienza una nueva partida
  {
    pos_bombero=0;
    bascula_tiempo=0;
    tick=0;
    marcatiempo=0;      // bascula que va asignando a cada nueva victima un intervalo de tiempo alternativo.
    vidas_restantes=0;
    puntos=0;
    ind_sig_actor=0;    // indice al vector "siguiente"
    izq_pres=der_pres=0;    // flags que indican si aun se tiene pulsada la tecla IZQ o DER, para evitar autorrepeticion
    retardo_victimas=0;
    retardo_maximo=RETARDO_TOTAL; // cuanto menor sea este nmero, ms rpido va el juego.

    borrar_sprites();
    init_victimas();

    sig_actor=siguiente[ind_sig_actor++];
    dibujar_bombero(pos_bombero,96);
    poner_puntos(puntos);

    while (vidas_restantes<3)  // BUCLE PRINCIPAL de juego: se ejecuta mientras an queden vidas
    {
      if (inkey(3,1) && !izq_pres && pos_bombero!=0)  // chequeamos que se pulse IZQ
      {
        dibujar_bombero(pos_bombero,100);
        pos_bombero--;
        dibujar_bombero(pos_bombero,96);
        izq_pres=1;
      }
      else if (inkey(4,1) && !der_pres && pos_bombero!=2) // chequeamos que se pulse DER
      {
        dibujar_bombero(pos_bombero,100);
        pos_bombero++;
        dibujar_bombero(pos_bombero,96);
        der_pres=1;
      }
      if (!inkey(3,1))  // si alguna se solto, reseteamos banderas
        izq_pres=0;
      if (!inkey(4,1))
        der_pres=0;

      for (actor=0;actor<10;actor++)  // y ahora comprobamos si la camilla ha tocado a alguna victima
      {
        if (VictimaActiva[actor])
        {
          pos_victima=VictimaPosicion[actor];
          esta_salvada=VictimaSalvada[actor];
          if (!esta_salvada && ((pos_victima==5 && pos_bombero==0) || (pos_victima==13 && pos_bombero==1) || (pos_victima==19 && pos_bombero==2)))
          {
            puntos++;
            poner_puntos(puntos);
            beep (50,20);
            VictimaSalvada[actor]=1;
          }
        }
      }

      retardo_victimas++;
      if (retardo_victimas>=retardo_maximo)  // esta parte se ejecuta slo cuando se alcanza el valor de retardo_maximo. Aqui es donde se mueven las victimas
      {
        retardo_victimas=0;
        bascula_tiempo=~bascula_tiempo;
        primera_victima_pintada=1;
        for (actor=0;actor<10;actor++)
        {
          if (!VictimaActiva[actor] || bascula_tiempo!=VictimaTick[actor])  // si la victima no est activa, o no le toca actualizarse, pasamos a la siguiente
            continue;

          pos_victima=VictimaPosicion[actor];
          esta_salvada=VictimaSalvada[actor];

          if (pos_victima>0)
            dibujar_victima(pos_victima,100);  // borramos victima de su antigua posicion

          pos_victima++;           // actualizamos posicion
          if (pos_victima==23)  // la posicion 23 es "dentro" de la ambulancia: victima fuera de pantalla
          {
            VictimaActiva[actor]=0;
            continue;
          }
          VictimaPosicion[actor]=pos_victima;

          if (primera_victima_pintada)  // solo hacemos "tick" una vez en cada sesion de movimiento.
          {
            beep(5,50);
            primera_victima_pintada=0;
          }
          dibujar_victima(pos_victima,96);  // dibujamos victima en su nueva ubicacin.

          if (pos_victima==6)   // este IF (y los dos siguientes que son iguales) comprueban si la victima est a punto de caerse y no ha tocado camilla. Si es as, chocazo que te cri.
          {
            if (!esta_salvada)
            {
              VictimaPosicion[actor]=0;
              pegarse_chocazo(0,pos_victima,&vidas_restantes);
            }
            else
              VictimaSalvada[actor]=0;
          }
          else if (pos_victima==14)
          {
            if (!esta_salvada)
            {
              VictimaPosicion[actor]=0;
              pegarse_chocazo(1,pos_victima,&vidas_restantes);
            }
            else
              VictimaSalvada[actor]=0;
          }
          else if (pos_victima==20)
          {
            if (!esta_salvada)
            {
              VictimaPosicion[actor]=0;
              pegarse_chocazo(2,pos_victima,&vidas_restantes);
            }
            else
              VictimaSalvada[actor]=0;
          }
        }

        if (bascula_tiempo)
        {
          tick++;        // tick cuenta los "ticks" que han de pasar hasta que le toca aparecer a la siguiente victima
          if (tick>=sig_actor)  // si le toca ya...
          {
            tick=0;
            retardo_maximo=(retardo_maximo<=75)? 75 : retardo_maximo-5;  // a medida que transcurre el juego, ste se hace ms rpido.
            sig_actor=siguiente[ind_sig_actor++];  // la siguiente victima aparecer dentro de "sig_actor" ticks
            if (sig_actor==255)
            {
              ind_sig_actor=0;
              sig_actor=siguiente[ind_sig_actor++];
              retardo_maximo=RETARDO_TOTAL/2;  // un pequeo respiro para que el juego no sea tan frentico.
            }
            for (actor=0;actor<10;actor++)  // busco un hueco en mi lista de vctimas inactivas, para poner a esta nueva vctima.
              if (!VictimaActiva[actor])
              {
                VictimaActiva[actor]=1;
                VictimaTick[actor]=marcatiempo;
                VictimaPosicion[actor]=0;
                VictimaSalvada[actor]=0;
                break;
              }
            marcatiempo=~marcatiempo;
          }
        }
      } //del IF de actualizar
    } //del bucle de juego de una vida
  } //del bucle principal de partida
}

void init_victimas (void)
{
  BYTE i;

  for (i=0;i<10;i++)
  {
    VictimaActiva[i]=0;
    VictimaPosicion[i]=0;
    VictimaTick[i]=0;
    VictimaSalvada[i]=0;
  }
}

void init_matrices (void)
{
  memcpy (posangelitos,"\x2\x16\x2\x19\x2\x1c",6);
  memcpy (poschocazos,"\x14\x08\x14\x0f\x14\x16",6);
  memcpy (posbombero,"\x11\x6\x11\xd\x11\x14",6);
  memcpy (posvictimas,"\x0a\x03\x05\x03\x08\x05\x0b\x06\x0e\x07\x11\x08\x0e\x09\x0b\x0a\x08\x0b\x05\x0c\x08\x0d\x0b\x0e\x0e\x0e\x11\x0f\x0e\x10\x0b\x11\x08\x12\x0b\x14\x0e\x15\x11\x16\x0e\x17\x0b\x19\x0e\x1b",46);
  memcpy (siguiente,"\x1\x16\xa\x9\x9\x9\x8\x8\x8\x7\x7\x7\x6\x6\x5\x5\x4\x4\x4\x4\x4\x3\x3\x3\x3\x3\x7\x6\x6\x6\x6\x6\x6\x6\x6\x6\x6\x6\x6\x5\x5\x5\x5\x5\x5\x5\x5\x5\x5\x4\x4\x4\x4\x4\x4\x4\x4\x4\x4\x14\x9\x9\x9\x9\x9\x8\x8\x8\x8\x8\x8\x8\x7\x7\x7\x7\x7\x7\x7\x7\x7\x7\x7\x7\x7\x7\x6\x6\x6\x6\x6\x6\x5\x5\x5\x5\x5\x5\x5\x5\x5\x5\x5\x4\x4\x4\x4\x4\x4\x4\xa\x3\x3\x3\x3\x3\x3\x3\x16\x1\x1\x1\x1\x1\x1\x1\x1\x1\x1\x1\x1\xff",132);
}

void dibujar_bombero (BYTE pos, BYTE color)
{
  BYTE *memattr;

  memattr=(BYTE *)(22528+posbombero[pos][0]*32+posbombero[pos][1]);
  memattr[0]=color;
  memattr[1]=color;
  memattr[4]=color;
  memattr[5]=color;
  memattr[32]=color;
  memattr[33]=color;
  memattr[36]=color;
  memattr[37]=color;
  memattr[64]=color;
  memattr[65]=color;
  memattr[66]=color;
  memattr[67]=color;
  memattr[68]=color;
  memattr[69]=color;
}

void dibujar_victima (BYTE pos, BYTE color)
{
  BYTE *memattr;

  memattr=(BYTE *)(22528+posvictimas[pos][0]*32+posvictimas[pos][1]);
  memattr[0]=color;
  memattr[1]=color;
  memattr[32]=color;
  memattr[33]=color;
}

void dibujar_angelito (BYTE pos, BYTE color)
{
  BYTE *memattr;

  memattr=(BYTE *)(22528+posangelitos[pos][0]*32+posangelitos[pos][1]);
  memattr[0]=color;
  memattr[1]=color;
  memattr[32]=color;
  memattr[33]=color;
  memattr[64]=color;
  memattr[65]=color;
}

void dibujar_chocazo (BYTE pos, BYTE color)
{
  BYTE *memattr;

  memattr=(BYTE *)(22528+poschocazos[pos][0]*32+poschocazos[pos][1]);
  memattr[0]=color;
  memattr[1]=color;
  memattr[31]=color;
  memattr[32]=color;
  memattr[33]=color;
  memattr[34]=color;
}

void pegarse_chocazo (BYTE poschoc, BYTE pos_victima, BYTE *vidas)
{
  BYTE i;

  dibujar_victima(pos_victima,100);
  dibujar_chocazo(poschoc,88);
  beep(250,10);
  (*vidas)++;
  if (*vidas==3)
  {
    pause(10);
    for (i=0;i<5;i++)
    {
      beep(250,10);
      pause(1);
    }
    pause(25);
  }
  else
  {
    pause(10); beep(500,10);
    pause(25);
  }
  dibujar_angelito((*vidas)-1,96);
  dibujar_chocazo(poschoc,91);
  if ((*vidas)==3)
  {
    locate(10,8);
    puts("\x4\xc7G A M E  O V E R");
    while (!inkey(5,2));
    locate(10,8);
    puts("\x4\x64                ");
  }
}

void poner_puntos (WORD puntos)
{
  BYTE numero[7];

  numero[0]=6;
  u16todec(puntos,numero+1);
  SETCOLOR(BRIGHT|PGREEN|IBLUE);
  locate(2,13);
  puts(numero);
}

void borrar_sprites (void)
{
  BYTE i;

  for (i=0;i<3;i++)
  {
    dibujar_bombero(i,100);
    dibujar_angelito(i,100);
    dibujar_chocazo(i,91);
  }
  for (i=0;i<23;i++)
    dibujar_victima(i,100);
}

void presentacion (void)
{
  SETCOLOR(BRIGHT|PGREEN|IBLACK);
  locate(7,6);
  puts("\x6\x7f 2008 McLeod/IdeaFix");

  SETCOLOR(FLASH|BRIGHT|PYELLOW|IRED);
  locate(10,10); puts("\x6 F I R E !! ");

  SETCOLOR(BRIGHT|PBLACK|IGREEN);
  locate(13,5); puts("Q-izquierda  P-derecha");

  SETCOLOR(BRIGHT|PBLACK|IGREEN);
  locate(16,5); puts("\x6Pulse S para comenzar");

  while(!inkey(5,2));

  SETCOLOR(BRIGHT|PGREEN|IGREEN);
  locate(7,6); puts("                     ");
  locate(10,10); puts("            ");
  locate(13,5); puts("                      ");
  locate(16,5); puts("                     ");
}
/* --------------------------------------------------------------------------------- */
/* --------------------------------------------------------------------------------- */
/* --------------------------------------------------------------------------------- */
#pragma disable_warning 85
int abs (int n)
{
  return (n>=0)?n:-n;
}

signed char sgn (int n)
{
  return (n>=0)?1:-1;
}

long frames (void) __naked
{
  _asm
  ld a,(#23672)
  ld l,a
  ld a,(#23673)
  ld h,a
  ld a,(#23674)
  ld e,a
  ld d,#0
  ret
  _endasm;
}

void pause (BYTE frames)
{
  BYTE i;

  for (i=0;i!=frames;i++)
  {
    WAIT_VRETRACE;
  }
}

void memset (BYTE *dir, BYTE val, WORD nby)
{
  _asm
  push bc
  push de
  ld l,4(ix)
  ld h,5(ix)
  ld a,6(ix)
  ld c,7(ix)
  ld b,8(ix)
  ld d,h
  ld e,l
  inc de
  dec bc
  ld (hl),a
  ldir
  pop de
  pop bc
  _endasm;
}

void memcpy (BYTE *dst, BYTE *fue, WORD nby)
{
  _asm
  push bc
  push de
  ld e,4(ix)
  ld d,5(ix)
  ld l,6(ix)
  ld h,7(ix)
  ld c,8(ix)
  ld b,9(ix)
  ldir
  pop de
  pop bc
  _endasm;
}

void cls (BYTE attr)
{
#ifdef USEROM
  _asm
  push bc
  push de
  ld a,4(ix)
  ld (#ATTRP),a
  call #0x0d6b
  ld a,#0xfe
  call #0x1601
  pop de
  pop bc
  _endasm;
#else
  memset((BYTE *)16384,0,6144);
  memset((BYTE *)22528,attr,768);
  SETCOLOR(attr);
  *((WORD *)SPOSN)=0;
#endif
}

void border (BYTE b)
{
  b&=0x7;
  ULA=b;
  *((BYTE *)BORDR)=b;
}

void puts (BYTE *str)
{
  volatile BYTE over=0;
  volatile BYTE bold=0;
  _asm
  push bc
  push de
  ld l,4(ix)
  ld h,5(ix)
buc_print:
  ld a,(hl)
  or a
  jr nz,no_fin_print
  jp fin_print
no_fin_print:
  cp #22
  jr nz,no_at
  inc hl
  ld a,(hl)
  ld (#ROW),a
  inc hl
  ld a,(hl)
  ld (#COLUMN),a
  inc hl
  jr buc_print
no_at:
  cp #13
  jr nz,no_cr
  xor a
  ld (#COLUMN),a
  ld a,(#ROW)
  inc a
  ld (#ROW),a
  inc hl
  jr buc_print
no_cr:
  cp #4
  jr nz,no_attr
  inc hl
  ld a,(hl)
  ld (#ATTRP),a
  inc hl
  jr buc_print
no_attr:
  cp #5
  jr nz,no_pr_over
  ld -1(ix),#0xff
  inc hl
  jr buc_print
no_pr_over:
  cp #6
  jr nz,no_pr_bold
  ld -2(ix),#0xff
  inc hl
  jr buc_print
no_pr_bold:
  cp #32
  jr nc,imprimible
  ld a,#32
imprimible:
  push hl
  ld hl,(#COLUMN)
  push hl
  push af
  ld de,#16384
  add hl,de
  ld a,h
  and #7
  rrca
  rrca
  rrca
  or l
  ld l,a
  ld a,#248
  and h
  ld h,a
  pop af
  push hl
  ld de,(#CHARS)
  ld l,a
  ld h,#0
  add hl,hl
  add hl,hl
  add hl,hl
  add hl,de
  pop de

  ld b,#8
print_car:
  ld a,(hl)
  sra a
  and -2(ix)
  or (hl)
  ld c,a
  ld a,(de)
  and -1(ix)
  xor c
  ld (de),a
  inc hl
  inc d
  djnz print_car

  pop hl
  push hl
  ld de,#22528
  ld b,h
  ld h,#0
  add hl,de
  xor a
  or b
  jr z,fin_ca_attr
  ld de,#32
calc_dirat:
  add hl,de
  djnz calc_dirat
fin_ca_attr:
  ld a,(#ATTRP)
  ld (hl),a
  pop hl
  inc l
  bit 5,l
  jr z,no_inc_fila
  res 5,l
  inc h
no_inc_fila:
  ld (#COLUMN),hl
  pop hl
  inc hl
  jp buc_print

fin_print:
  pop de
  pop bc
  _endasm;
}

void locate (BYTE row, BYTE col)
{
  *((BYTE *)ROW)=row;
  *((BYTE *)COLUMN)=col;
}

void putdec (WORD n)
{
  BYTE num[6];

  u16todec (n,num);
  puts (num);
}

void u16todec (WORD n, char *s)
{
  BYTE i=4;
  WORD divisor=10, resto;

  memset(s,' ',5);
  do
  {
    resto=n%divisor;
    n/=divisor;
    s[i--]=resto+'0';
  }
  while (n);
  s[5]='\0';
}

void u16tohex (WORD n, char *s)
{
  u8tohex((n>>8)&0xFF,s);
  u8tohex(n&0xFF,s+2);
}

void u8tohex (BYTE n, char *s)
{
  BYTE i=1;
  BYTE resto;

  resto=n&0xF;
  s[1]=(resto>9)?resto+55:resto+48;
  resto=n>>4;
  s[0]=(resto>9)?resto+55:resto+48;
  s[2]='\0';
}

void wait_key (void)
{
  while ((ULA&0x1f)==0x1f);
}

BYTE inkey (BYTE semif, BYTE pos)
{
  BYTE teclas;
  BYTE posbit;

  switch (semif)
  {
    case 1: teclas=SEMIFILA1; break;
    case 2: teclas=SEMIFILA2; break;
    case 3: teclas=SEMIFILA3; break;
    case 4: teclas=SEMIFILA4; break;
    case 5: teclas=SEMIFILA5; break;
    case 6: teclas=SEMIFILA6; break;
    case 7: teclas=SEMIFILA7; break;
    case 8: teclas=SEMIFILA8; break;
    default: teclas=ULA; break;
  }
  posbit=1<<(pos-1);
  return (teclas&posbit)?0:1;
}

void beep (WORD durmili, BYTE freq) __critical
{
  volatile BYTE cborde;

  cborde=(*(BYTE *)(BORDR));
  _asm
  push bc
  push de

  ld l,6(ix)  ;se desplaza dos byte por ser "critical".
  ld h,7(ix)
  ld d,-1(ix)

  ld b,8(ix)
  xor a
  sub b
  ld b,a
  ld c,a

bucbeep:
  ld a,d
  xor #0x18
  ld d,a
  out (#254),a

  ld b,c
bucperiodobeep:
  djnz bucperiodobeep

  dec hl
  ld a,h
  or l
  jr nz,bucbeep

  pop de
  pop bc
  _endasm;
}
